/***************************************************************************
*                                                                          *
*   hardware.c:     Linux Device Driver for the astronomical CCD camera    *
*                   LcCCD14 by OES GmbH                                    *
*                                                                          *
*                   Low level hardware functions                           *
*                                                                          *
*   Author:         Robert Schwebel (robert@schwebel.de)                   *
*                                                                          *
*   Revision:       0.1.0 (19.09.1999)                                     *
*                                                                          *
****************************************************************************
*                                                                          *
* old bugs, perhaps out of order:                                          *
*                                                                          *
*	     - variables should be static?				   *
*            - docs for LcCCD14_Reset are missing (ask Doc Fleischmann!)   *
*            - undocumented start_pulse in command writing sequence!       *
*            - #ifdefs are not working properly (LcCCD14_Transfer)         *
*            - CCDLIB_PictureparamStruct is not yet understood properly    *
*                                                                          *
* TODO:      - rewrite port handling macros with functions                 *
*            - port access is hardwired!                                   *
*            - define a function LcCCD14_TestUserBreak()                   *
*            - rename LcCCD14Setup_Params constants to english names       *
*                                                                          *
***************************************************************************/


/* ----------------------------------------------------------------------- */
/*                                                                         */
/* Include some .h files                                                   */
/*                                                                         */
/* ----------------------------------------------------------------------- */

#include "hardware.h"                        /* our own declarations       */
#include <stdio.h>			     /* define the NULL pointer    */
#include <asm/io.h>                          /* port input output          */
#include <linux/delay.h>                     /* time delay loop            */


/* ----------------------------------------------------------------------- */
/*                                                                         */
/* Internal Macros: basic definitions                                      */
/*                                                                         */
/* ----------------------------------------------------------------------- */

/* These macros convert the temperature values                             */
#define	M                        1.3384
#define	T                      -17.2548

#define temp2val(Temp)         (int)(((double)(Temp)-T)/M+.5)
#define val2temp(Val)          (int)(M*(double)(Val)+T)

/* base addresses for printer ports                                        */
#define LPT1_Address	       0x378
#define LPT2_Address	       0x278

/* offset of printer port registers, relative to base address              */
#define LPT_DataReg            0x00
#define LPT_StatReg            0x01
#define LPT_CommReg            0x02

/* camera control commands                                                 */
#define LcCCD14_CMD_NOP        0x00
#define LcCCD14_CMD_READPIXEL  0x01
#define LcCCD14_CMD_RESET_AB   0x02
#define LcCCD14_CMD_TRANS_A_B  0x03
#define LcCCD14_CMD_TRANS_AB_C 0x04
#define LcCCD14_CMD_SETUP      0x05


/* ----------------------------------------------------------------------- */
/*                                                                         */
/* These variables mirror the printer port registers.                      */
/* Initialized with standard values.                                       */
/*                                                                         */
/* VARIABLES DEFINED HERE ARE GLOBAL!                                      */
/*                                                                         */
/* ----------------------------------------------------------------------- */

static unsigned char DataRegVal = 0xff;      /* telescope drive-control    */
                                             /* pins are initially set to  */
					     /* HI                         */
static unsigned char StatRegVal = 0x00;
static unsigned char CommRegVal = 0x00;
static unsigned int Port = LPT1_Address;     /* default port is LPT1       */

static unsigned int  uiTimecount;
static unsigned int  uiTimeout=10000;        /* timeout limit              */
static unsigned char uchTimeoutreached=0;    /* timeout flag set?          */
 

/* ----------------------------------------------------------------------- */
/*                                                                         */
/* Internal Macros: Low level port handling                                */
/*                                                                         */
/* DataRegVal (port data register) Bit definition:                         */
/*                                                                         */
/* bit 4: telescope drive control: left                                    */
/* bit 5: telescope drive control: left                                    */
/* bit 6: telescope drive control: left                                    */
/* bit 7: telescope drive control: left                                    */
/*                                                                         */
/* ----------------------------------------------------------------------- */

/* telescope drive control pin: right                                      */
#define  right_low  DataRegVal&=0xdf;outb(DataRegVal,Port+LPT_DataReg);
#define  right_high DataRegVal|=0x20;outb(DataRegVal,Port+LPT_DataReg);

/* telescope drive control pin: left                                       */
#define  left_low   DataRegVal&=0xef;outb(DataRegVal,Port+LPT_DataReg);
#define  left_high  DataRegVal|=0x10;outb(DataRegVal,Port+LPT_DataReg);

/* telescope drive control pin: up                                         */
#define  up_low     DataRegVal&=0xbf;outb(DataRegVal,Port+LPT_DataReg);
#define  up_high    DataRegVal|=0x40;outb(DataRegVal,Port+LPT_DataReg);

/* telescope drive control pin: down                                       */
#define  down_low   DataRegVal&=0x7f;outb(DataRegVal,Port+LPT_DataReg);
#define  down_high  DataRegVal|=0x80;outb(DataRegVal,Port+LPT_DataReg);

/* START pin (inverse logic: low=active)                                   */
#define  start_low  CommRegVal|=0x01;outb(CommRegVal,Port+LPT_CommReg);
#define  start_high CommRegVal&=0xfe;outb(CommRegVal,Port+LPT_CommReg);

/* MODE pin (non inverse logic: high=active)                               */
#define  mode_low   DataRegVal&=0xfe;outb(DataRegVal,Port+LPT_DataReg);
#define  mode_high  DataRegVal|=0x01;outb(DataRegVal,Port+LPT_DataReg);

/* Wait for negative edge on PS line                                       */
/* sets uchTimeoutreached if a timeout occured                             */

#define wait_pslow                                                         \
  uiTimecount=uiTimeout;                                                   \
  while(((inb(Port+LPT_StatReg)&0x80)==0)&&(uiTimecount>0)){               \
    uiTimecount--;                                                         \
  };                                                                       \
  if(uiTimecount==0){uchTimeoutreached|=0x01;};

/* Wait for positive edge on PS line                                       */
/* sets uchTimeoutreached if a timeout occured                             */

#define wait_pshigh                                                        \
  uiTimecount=uiTimeout;                                                   \
  while(((inb(Port+LPT_StatReg)&0x80)!=0)&&(uiTimecount>0)){               \
    uiTimecount--;                                                         \
  };                                                                       \
  if (uiTimecount==0){uchTimeoutreached|=0x01;};

/* Read value from camera                                                  */
#define get_value(uchValue)                                                \
  uchValue = (inb(Port+LPT_StatReg)>>3)&0x0f ;

/* Read value from camera at low edge of PS signal                         */
#define L_get_value(uchValue)                                              \
  wait_pslow;                                                              \
  uchValue = (unsigned char)(inb(Port+LPT_StatReg)&0x78);                  \
  start_low;

/* Read value from camera at high edge of PS signal                        */
#define H_get_value(uchValue)                                              \
  wait_pshigh;                                                             \
  uchValue = (unsigned char)(inb(Port+LPT_StatReg)&0x78);                  \
  start_high;

/* Readout status register                                                 */
#define readLPTstat(val)                                                   \
 val=inb(Port+LPT_StatReg) & 0x80;


/* ----------------------------------------------------------------------- */
/*                                                                         */
/* Internal Macros which are not directly hardware dependend               */
/*                                                                         */
/* ----------------------------------------------------------------------- */

/* low pulse on start line                                                 */
#define	start_puls                                                         \
  start_low;                                                               \
  wait_pslow;                                                              \
  start_high;                                                              \
  wait_pshigh;

/* low pulse on mode line                                                  */
#define	mode_puls                                                          \
  mode_low;                                                                \
  wait_pslow;                                                              \
  mode_high;                                                               \
  wait_pshigh;


/* ----------------------------------------------------------------------- */
/*                                                                         */
/* Internal Functions                                                      */
/*                                                                         */
/* ----------------------------------------------------------------------  */

/***************************************************************************
*                                                                          *
*   LcCCD14_Command:   Execute camera command.                             *
*                                                                          *
*   Paramterers:    uchCommand                                             *
*                   Command that is to be executed.                        *
*                                                                          *
*   Return Value:   LcCCD14_Successful: everything ok                      *
*                   LcCCD14_Timeout:    camera does not react              *
*                                                                          *
*   Author:         Robert Schwebel (r.schwebel@tu-bs.de)                  *
*                                                                          *
*   Revision:       1.0 (07.09.1994)                                       *
*                                                                          *
***************************************************************************/

enum LcCCD14_ErrorEnum LcCCD14_Command(unsigned char uchCommand)
{
  unsigned int iCount; 

  for(iCount=0; iCount<=uchCommand; iCount++)
    {mode_puls;};
  start_puls;                           /* start a camera command         */
  return(LcCCD14_Successful);
}


/***************************************************************************
*                                                                          *
*   LcCCD14_GetVal:   Read value from camera                               *
*                                                                          *
*   Parameters:     *uchValue                                              *
*                   Readout value from camera.                             *
*                                                                          *
*   Return Value:   LcCCD14_Successful: everything ok                      *
*                   LcCCD14_Timeout:    camera does not react              *
*                                                                          *
*   Revision:       1.0 (07.09.1994)                                       *
*                                                                          *
***************************************************************************/

enum LcCCD14_ErrorEnum LcCCD14_GetVal(unsigned char *uchValue)
{
  wait_pslow;                          /* wait for camera to send nibble   */
  get_value(*uchValue);                /* read nibble                      */
  start_low;                           /* do handshake                     */
  wait_pshigh;
  start_high;

  return(LcCCD14_Successful);
}


/* ----------------------------------------------------------------------- */
/*                                                                         */
/* API Functions                                                           */
/*                                                                         */
/* ----------------------------------------------------------------------- */

/***************************************************************************
*                                                                          *
*   LcCCD14_Reset:  Reset camera to initial state.                         *
*                                                                          *
*   Parameters:     None.                                                  *
*                                                                          *
*   Return Value:   LcCCD14_Successful: everything ok                      *
*                   LcCCD14_ResetError: reset was not successful           *
*									   *
*   ATTENTION:      Does direct hardware access!                           *
*                                                                          *
*   Revision:       1.0 (RS, 12.02.1995)                                   *                                                                           *
*                                                                          *
***************************************************************************/

enum LcCCD14_ErrorEnum LcCCD14_Reset(void)
{
  long i;
  int j;
  int l;
  int k = 0;
  int m = 1500;

  do{
    k++; m--;
    mode_low;
    for(i=0x00,l=1000;(l>0)&&(i==0);l--){
      i=(inb(Port+LPT_StatReg)&0x80);
    }
    mode_high;
    for(j=0x80,l=1000;(l>0)&&(j!=0);l--){
      j=(inb(Port+LPT_StatReg)&0x80);
    }
    if ((i==0) || (j!=0)) {k = 0;};
    start_low;
    for(i=0x00,l=1000;(l>0)&&(i==0);l--){
      i=(inb(Port+LPT_StatReg)&0x80);
    }
    start_high;
    for(j=0x80,l=1000;(l>0)&&(j!=0);l--){
      j=(inb(Port+LPT_StatReg)&0x80);
    }
    if ((i==0) || (j!=0)) {k = 0;};
  } while((k < 10) && (m>0));
  
  if (m == 0) return(LcCCD14_ResetError);

  return(LcCCD14_Successful);
}


/***************************************************************************
*                                                                          *
*   LcCCD14_Setup:   Send setup parameters to camera.                      *
*                                                                          *
*   Parameters:     *parameters                                            *
*                   Setup parameters to send                               *
*                                                                          *
*   Return Value:   LcCCD14_Successful: everything ok                      *
*                   LcCCD14_Timeout:    camera does not react              *
*                   LcCCD14_EchoErr:    line echo test error               *
*                   LcCCD14_WrongParam: wrong parameters                   *
*                                                                          *
*   Revision:       1.0 (RS, 07.09.1994)                                   *
*                                                                          *
***************************************************************************/

enum LcCCD14_ErrorEnum LcCCD14_Setup(LcCCD14_SetupParams *parameters)
{
  unsigned char uchEcho;
  unsigned char uchTemp1;               /* temporary variable              */
  unsigned char uchTemp2;               /* dto.                            */
  unsigned char	uchControl;             /* control byte                    */

  /* Description of control byte:                                          */
        
  /* Bit 0: Binning0                    0: on     1: off                   */
  /* Bit 1: (Binning1)                  0:        1:                       */
  /* Bit 2: 0                           0:        1:                       */
  /* Bit 3: Frame Number                0: 1      1: 0                     */
  /* Bit 4: AD-Resolution               0: 4Bit   1: 12 Bit                */
  /* Bit 5: Correlated Doublesampling   0: on     1: off                   */
  /* Bit 6: Mechanical Shutter (opt.)   0: open   1: closed                */
  /* Bit 7: ?                                                              */

  /* set global timeout pointer to zero                                    */
  uchTimeoutreached = 0;

  /* construct camera control byte                                         */
  uchControl = 0;
  switch ((*parameters).iBinning){
    case einBinning: uchControl |= 0x01;
		     break;
    case ausBinning: break;

    default: return(LcCCD14_WrongParam);
  }
  switch ((*parameters).iADResolution){
    case AD12Bit:    uchControl |= 0x10;
		     break;
    case AD4Bit:     break;

    default: return(LcCCD14_WrongParam);
  }
  switch ((*parameters).iSampling){
    case korellAbt:  uchControl |= 0x20;
		     break;
    case einfachAbt: break;

    default: return(LcCCD14_WrongParam);
  }

  LcCCD14_Command(LcCCD14_CMD_SETUP);     /* send setup command            */
  start_puls;				  /* empirically... ???            */

  LcCCD14_GetVal(&uchTemp1); 		  /* get hardware version number   */
					  /* ... not used!                 */
  LcCCD14_Command(0x0F);        	  /* line test pattern             */
  LcCCD14_GetVal(&uchEcho);               /* ... gets echoed by camera     */

  LcCCD14_Command(uchControl);		  /* send camera control byte      */
  LcCCD14_Command(1);			  /* no left border                */
  if ((*parameters).iBinning==einBinning) /* imaging area:                 */
    {LcCCD14_Command(392/8+1);}        	  /*  392 pixel with binning       */
  else {LcCCD14_Command(784/8+1);}	  /*  784 pixel without binning    */
  LcCCD14_Command(1);			  /* no right border               */

  /* send desired temperature                                              */
  LcCCD14_Command(temp2val((double)((*parameters).iDesiredTemp)));

  /* read current temperature                                              */
  LcCCD14_GetVal(&uchTemp1);
  LcCCD14_GetVal(&uchTemp2);
  (*parameters).iMeasuredTemp = val2temp((uchTemp1<<4) + (uchTemp2));

  if (uchTimeoutreached!=0){ return(LcCCD14_Timeout);}
  if (uchEcho!=0x0F){return(LcCCD14_EchoErr);}
  return(LcCCD14_Successful);
}


/***************************************************************************
*                                                                          *
*   LcCCD14_Integrate:                                                     *
*                                                                          *
*                   Integrate image                                        *
*                                                                          *
*   Parameters:     fInttime                                               *
*                   integration time                                       *                             *
*                                                                          *
*   Result:         LcCCD14_Successful: everything ok                      *
*                   LcCCD14_Timeout:    camera does not react              *
*                   LcCCD14_Userbreak:  user pressed escape                *
*                   LcCCD14_WrongParam: wrong parameters                   *
*                                                                          *
*   Revision:       1.0 (RS, 07.09.1994)                                   *
*                                                                          *
***************************************************************************/

enum LcCCD14_ErrorEnum LcCCD14_Integrate(float fInttime)
{
  int          iUserBreak = 0;
  unsigned int uiCount;

  if ((fInttime<0) || (fInttime>3600)) {return(LcCCD14_WrongParam);};

  uchTimeoutreached = 0;

  /* command: clear image area                                             */
  LcCCD14_Command(LcCCD14_CMD_RESET_AB);
  start_puls;

  for (uiCount=0; 
       (uiCount<(int)(fInttime/0.1))&&(iUserBreak==0); 
       uiCount++)
  {
/*     iEscape=LcCCD14_TestUserBreak(); */
    udelay(100000);                          /* integrate desired time     */
  }
  udelay((long)((long)(fInttime*1000l)%100000l)); /* Rest                  */

  /* command: transfer image zone to dark zone                             */
  LcCCD14_Command(LcCCD14_CMD_TRANS_A_B);
  start_puls;

  if (iUserBreak==1) {return( LcCCD14_Userbreak ); }
  if (uchTimeoutreached!=0){return(LcCCD14_Timeout);}
  return(LcCCD14_Successful);
}


/***************************************************************************
*                                                                          *
*   LcCCD14_Prepare_Transfer:                                              *
*                                                                          *
*                    Before transferring rows we have to read the first    *
*                    four rows (which contain no useful information)       *
*                                                                          *
*   Parameters:      None                                                  *
*                                                                          *
*   Result:          LcCCD14_Successful: everything ok                     *
*                                                                          *
*   Revision:        1.0 (RS, 02.04.1999)                                  *
*                                                                          *
***************************************************************************/

enum LcCCD14_ErrorEnum LcCCD14_Prepare_Transfer(void)
{

  int iY;

  for (iY=4; iY!=0; iY--){	             /* 1st 4 lines contain nothing*/
  
    LcCCD14_Command(LcCCD14_CMD_TRANS_AB_C);    
    start_puls;				     /* ???                        */

  }
  
  return(LcCCD14_Successful);
}


/***************************************************************************
*                                                                          *
*   LcCCD14_Transfer_Row:                                                  *
*                                                                          *
*                    Transfer one image row. We can only transfer one      *
*                    row at once due to kernel kmalloc restrictions.       * 
*                                                                          *
*   Parameters:      **ppBuffer:   the pixel values are transferred to     *
*                                  this buffer. This buffer has to have a  *
*                                  length of 2048 bytes!!!                 *
*                    *pParameters: camera setup parameters                 *
*                    *count:       number of bytes(!) transferred          *
*                                                                          *
*   Result:          LcCCD14_Successful: everything ok                     *
*                    LcCCD14_Timeout:    camera does not react             *
*                    LcCCD14_Userbreak:  user break                        *
*                    LcCCD14_OutOfMem:   no memory left                    *
*                    LcCCD14_WrongParam: wrong parameters                  *
*                                                                          *
*   Revision:        0.2 (RS, 15.09.1999)                                  *
*                    0.1 (02.04.1999)                                      *
*                                                                          *
***************************************************************************/

enum LcCCD14_ErrorEnum LcCCD14_Transfer_Row(
                 __u8                      **ppBuffer,
		 LcCCD14_SetupParams       *pParameters,
		 int                       *count
){


  int            iRowLength;
  int		 iX;         /* Pixel counter                              */ 
  unsigned char  uchNib1;    /* One pixel has a maximum of three           */
  unsigned char  uchNib2;    /* 4bit nibbles.                              */
  unsigned char  uchNib3;         


  /* determine dimensions of the image from given camera parameters        */

  if (pParameters->iBinning==einBinning)     /* find out length of rows    */
    iRowLength=392;
  else 
    iRowLength=784;

  *count = iRowLength*sizeof(short);         /* Bytes per pixel            */

  /* start transfer                                                        */

  if (pParameters->iADResolution==AD4Bit){

    /* 4 bits per pixel                                                    */
    LcCCD14_Command(LcCCD14_CMD_TRANS_AB_C);    
    start_puls;	                                /* ???                     */
    LcCCD14_Command(LcCCD14_CMD_READPIXEL);     
    start_puls;                                 /* ???                     */

    for ( iX=1; iX<=iRowLength/2; iX++ ){
      /* read nibble from bits ".XXXX..." and write to memory              */
      L_get_value(uchNib1);
      H_get_value(uchNib2);
      **ppBuffer = (u8) (0x00);
      *ppBuffer+=sizeof(u8);
      **ppBuffer = (u8) (uchNib1>>3);
      *ppBuffer+=sizeof(u8);
      **ppBuffer = (u8) (0x00);
      *ppBuffer+=sizeof(u8);
      **ppBuffer = (u8) (uchNib2>>3);
      *ppBuffer+=sizeof(u8);
    }

  }

  else{

    /* 12 bits per pixel                                                   */
    LcCCD14_Command(LcCCD14_CMD_TRANS_AB_C);
    start_puls;	                                /* ???                     */
    LcCCD14_Command(LcCCD14_CMD_READPIXEL);   
    start_puls;                                 /* ???                     */

    for ( iX=0; iX<(iRowLength/2); iX++ ){
      /* read high nibble from bits ".XXXX..." and write to memory         */
      L_get_value(uchNib3);
      /* get mid and low nibble                                            */
      H_get_value(uchNib2);
      L_get_value(uchNib1);

      **ppBuffer = (u8)((uchNib2<<1)|(uchNib1>>3));
      *ppBuffer += sizeof(u8);
      **ppBuffer = (u8)(uchNib3>>3);
      *ppBuffer += sizeof(u8);

      /* get high nibble from bits ".XXXX..." and write to memory          */
      H_get_value(uchNib3);
      /* get mid and low nibble                                            */
      L_get_value(uchNib2);
      H_get_value(uchNib1);
      **ppBuffer = (u8)((uchNib2<<1)|(uchNib1>>3));
      *ppBuffer += sizeof(u8);
      **ppBuffer = (u8)(uchNib3<<3);
      *ppBuffer += sizeof(u8);
    }

  }

  return(LcCCD14_Successful); 
}
