/***************************************************************************
*                                                                          *
*   hardware.c:     Linux Device Driver for the astronomical CCD camera    *
*                   LcCCD14 by OES GmbH                                    *
*                                                                          *
*                   Low level hardware functions                           *
*                                                                          *
*   Author:         Robert Schwebel (robert@schwebel.de)                   *
*                                                                          *
*   Revision:       0.1.0 (19.09.1999)                                     *
*                                                                          *
***************************************************************************/

#ifndef __HARD14
#define __HARD14

/* ---------------------------------------------------------------------- */
/*                                                                        */
/* Define some variables                                                  */
/*                                                                        */
/* ---------------------------------------------------------------------- */

/* All functions return these results */
/* ---------------------------------- */
enum LcCCD14_ErrorEnum
{
  LcCCD14_Successful,
  LcCCD14_ResetError,
  LcCCD14_Timeout,
  LcCCD14_Userbreak,
  LcCCD14_OutOfMem,
  LcCCD14_WrongParam,
  LcCCD14_EchoErr,
  LcCCD14_Undef
};

/* Camera Parameters */
/* ----------------- */
typedef struct {
  int iIntegrationtime; /* in milliseconds */
  int iDesiredTemp;
  int iMeasuredTemp;
  int iCooling;
  int iADResolution;
  int iBinning;
  int iSampling;
  int iInterl;}
LcCCD14_SetupParams;

/* Image Information */
/* ----------------- */
typedef struct {
  long          lLineLength;                 /* number of pixels per line  */
  long          lColumnHeight;               /* number of pixels p. column */
  double        dXScaleDelta;                /* absolute size of pixel, x  */
  double        dYScaleDelta;}               /* absolute size of pixel, y  */
LcCCD14_ImageInformation;


#define einKuehlung 1                        /* Constants for camera setup */
#define ausKuehlung 0
#define AD12Bit     1
#define AD4Bit      0
#define einBinning  1
#define ausBinning  0
#define einfachAbt  0
#define korellAbt   1
#define einInterlc  1
#define ausInterlc  0

/* ---------------------------------------------------------------------- */
/*                                                                        */
/* Abstraction functions: API                                             */
/*                                                                        */
/* ---------------------------------------------------------------------- */

/***************************************************************************
*                                                                          *
*   LcCCD14_Reset:   Reset camera to initial state.                        *
*                                                                          *
*   Parameters:     None.                                                  *
*                                                                          *
*   Return Value:   LcCCD14_Successful: everything ok                      *
*                   LcCCD14_ResetError: reset was not successful           *
*									   *
*   ATTENTION:      Does direct hardware access!                           *
*                                                                          *
*   Revision:       1.0 (RS, 12.02.1995)                                   *                                                                           *
*                                                                          *
***************************************************************************/

enum LcCCD14_ErrorEnum LcCCD14_Reset(void);


/***************************************************************************
*                                                                          *
*   LcCCD14_Setup:   Send setup parameters to camera.                      *
*                                                                          *
*   Parameters:     *parameters                                            *
*                   Setup parameters to send                               *
*                                                                          *
*   Return Value:   LcCCD14_Successful: everything ok                      *
*                   LcCCD14_Timeout:    camera does not react              *
*                   LcCCD14_EchoErr:    line echo test error               *
*                   LcCCD14_WrongParam: wrong parameters                   *
*                                                                          *
*   Revision:       1.0 (RS, 07.09.1994)                                   *
*                                                                          *
***************************************************************************/

enum LcCCD14_ErrorEnum LcCCD14_Setup(LcCCD14_SetupParams *parameters);


/***************************************************************************
*                                                                          *
*   LcCCD14_Integrate:                                                     *
*                                                                          *
*                   Integrate image                                        *
*                                                                          *
*   Parameters:     fInttime                                               *
*                   integration time                                       *                             *
*                                                                          *
*   Result:         LcCCD14_Successful: everything ok                      *
*                   LcCCD14_Timeout:    camera does not react              *
*                   LcCCD14_Userbreak:  user pressed escape                *
*                   LcCCD14_WrongParam: wrong parameters                   *
*                                                                          *
*   Revision:       1.0 (RS, 07.09.1994)                                       *
*                                                                          *
***************************************************************************/

enum LcCCD14_ErrorEnum LcCCD14_Integrate(float fInttime);

/***************************************************************************
*                                                                          *
*   LcCCD14_Prepare_Transfer:                                              *
*                                                                          *
*                    Before transferring rows we have to read the first    *
*                    four rows (which contain no useful information)       *
*                                                                          *
*   Parameters:      None                                                  *
*                                                                          *
*   Result:          LcCCD14_Successful: everything ok                     *
*                                                                          *
*   Revision:        1.0 (RS, 02.04.1999)                                      *
*                                                                          *
***************************************************************************/

enum LcCCD14_ErrorEnum LcCCD14_Prepare_Transfer(void);

/***************************************************************************
*                                                                          *
*   LcCCD14_Transfer_Row:                                                  *
*                                                                          *
*                    Transfer one image row. We can only transfer one      *
*                    row at once due to kernel kmalloc restrictions.       * 
*                                                                          *
*   Parameters:      **ppBuffer:   the pixel values are transferred to     *
*                                  this buffer. This buffer has to have a  *
*                                  length of 2048 bytes!!!                 *
*                    *pParameters: camera setup parameters                 *
*                    *count:       number of bytes(!) transferred          *
*                                                                          *
*   Result:          LcCCD14_Successful: everything ok                     *
*                    LcCCD14_Timeout:    camera does not react             *
*                    LcCCD14_Userbreak:  user break                        *
*                    LcCCD14_OutOfMem:   no memory left                    *
*                    LcCCD14_WrongParam: wrong parameters                  *
*                                                                          *
*   Revision:        0.2 (RS, 15.09.1999)                                  *
*                    0.1 (02.04.1999)                                      *
*                                                                          *
***************************************************************************/

enum LcCCD14_ErrorEnum LcCCD14_Transfer_Row(
                 __u8                      **ppBuffer,
		 LcCCD14_SetupParams       *pParameters,
		 int                       *count
);


#endif
