/***************************************************************************
*                                                                          *
*   lcccd14.c:      Linux Device Driver for the astronomical CCD camera    *
*                   LcCCD14 by OES GmbH                                    *
*                                                                          *
*                   Based on the Linux Kernel Module Programming Guide     *
*                   by Ori Pomerantz. Parts of the hardware routines       *
*                   are derived from the original OES driver software by   *
*                   Matthias Pruksch.                                      *
*                                                                          *
*   Author:         Robert Schwebel (r.schwebel@tu-bs.de)                  *
*                                                                          *
*   Revision:       0.1.0 (19.09.1999)                                     *
*                                                                          *
*   Bugs:           - hide symbol table (->Rubini)?                        *
*                   - include sys/file.h ? types! -> compiler warnings     *
*                   - clear up all compiler warnings                       *
*                   - is it possible to read device numbers on runtime?    *
*                   - LcCCD14_DEBUG_REMOVE has to be tested                *
*                                                                          *
*   ToDo:           - dynamic allocation of maior device numbers           *
*                   - implement proper (/proc) status return facility      *
*                                                                          *
***************************************************************************/

/* Some definitions necessary for kernel modules ------------------------- */

#define __KERNEL__            /* kernel space programs have to exclude     */
                              /* some stuff in the standard include files  */

#define MODULE                /* we are writing a module here              */

/* Standard in kernel modules -------------------------------------------- */

#include <linux/kernel.h>     /* We're doing kernel work                   */
#include <linux/module.h>     /* Specifically, a module                    */
#include <linux/proc_fs.h>    /* want to use the proc filesystem           */

/* For character devices ------------------------------------------------- */

#include <linux/fs.h>         /* The character device definitions are here */
#include <linux/wrapper.h>    /* A wrapper which does next to nothing at   */
                              /* at present, but may help for compatibili- */
                              /* ty with future versions of Linux          */
#include <linux/ioport.h>     /* reserve ports                             */

/* register_symtab(NULL); */  /* hide symbol table                         */

/* Our own API for communication with the camera ------------------------- */

#include "hardware.h"         /* Interface to the LcCCD14 camera.          */
#include "hardware.c"         /* Complete source is included here!         */
/* #include <sys/file.h> */

#include "lcccd14.h"          /* API functions for camera handling         */


/* ----------------------------------------------------------------------- */
/*                                                                         */
/* Global Device Driver Variables                                          */
/*                                                                         */
/* ----------------------------------------------------------------------- */

/* The name for our device, as it will appear in /proc/devices             */
#define DEVICE_NAME "lcccd14"
#define DEVICE_MAJOR 34

/* The maximum length of the data from the device. One row contains 784    */
/* pixels max. As you can only get kernel memory in blocks of 2**N bytes   */
/* and the maximum buffer size in kernel space is 131056 bytes we reserve  */
/* memory only for one row here.                                           */
#define BUF_LEN 2048

/* Is the device open right now? Used to prevent concurent access into     */
/* the same device ("in-use counter")                                      */
static int iDevice_Open = 0;

/* Memory buffer in kernel space for image data                            */
static u8    u8RowBuffer[BUF_LEN]; /* buffer for image data                */
static u8    *pu8RowBuffer;
static int   iBytesInBuffer = 0;   /* bytes currently in the buffer        */
static int   iRowsLeft = 0;        /* how many rows wait for download?     */

/* Global Parameters with preferences for the next image aqquisition.      */
/* After each integration the temperature value can be read from here.     */
static LcCCD14_SetupParams SetupParams;


/***************************************************************************
*                                                                          *
*   device_open:    This function is called whenever a process attempts    *
*                   to open() the device file.                             *
*                                                                          *
*   Parameters:     inode, file:   this is convention for kernel modules   *
*                                                                          *
*   Return Value:   SUCCESS if successful :-)                              *
*                                                                          *
*   Revision:       0.0.2 (RS, 19.09.1999)                                 *                                         
*                                                                          *
***************************************************************************/

static int device_open(struct inode *inode, struct file *file) {

  unsigned int uiMinor = MINOR(inode->i_rdev);

  printk ("LcCCD14: opening device\n");

  /* Check if we have a valid lp number                                    */
  if (uiMinor >= LP_NO)
    return -ENXIO;

  /* We don't want to talk to two processes at the same time               */
  if (iDevice_Open)
    return -EBUSY;

  /* If this was a process, we would have had to be more careful here!     */
  /*                                                                       */
  /* In the case of processes, the danger would be that one process        */ 
  /* might have check Device_Open and then be replaced by the scheduler    */  
  /* by another process which runs this function. Then, when the first     */ 
  /* process was back on the CPU, it would assume the device is still not  */ 
  /* open. However, Linux guarantees that a process won't be replaced      */
  /* while it is running in kernel context.                                */
  /*                                                                       */
  /* In the case of SMP, one CPU might increment Device_Open while another */
  /* CPU is here, right after the check. However, in version 2.0 of the    */
  /* kernel this is not a problem because there's a lock to guarantee      */
  /* only one CPU will be kernel module at the same time. This is bad in   */
  /* terms of performance, so it will probably be changed in the future,   */
  /* but in a safe way.                                                    */

  iDevice_Open++;
  
  pu8RowBuffer = &u8RowBuffer[0];       /* Init pointer to RowBuffer       */    
  iRowsLeft = 0;                        /* actually no rows to read        */

  /* Make sure that the module isn't removed while the file is open by     */
  /* incrementing the usage count (the number of opened references to the  */
  /* module, if it's not zero rmmod will fail)                             */
  MOD_INC_USE_COUNT;

  return SUCCESS;

}


/***************************************************************************
*                                                                          *
*   device_release: This function is called when a process closes the      *
*                   device file. It is not allowed to fail.                *
*                                                                          *
*   Parameters:     inode, file:   this is convention for kernel modules   *
*                                                                          *
*   Return Value:   None.                                                  *
*                                                                          *
*   Revision:       0.0.1 (30.03.1999)                                     *                                         
*                                                                          *
***************************************************************************/

static void device_release(struct inode *inode, struct file *file)
{

  printk ("LcCCD14: closing device\n");

  /* We're now ready for our next caller                                   */
  iDevice_Open --;

  /* Decrement the usage count, otherwise once you opened the file you'll
   * never get rid of the module.
   */
  MOD_DEC_USE_COUNT;
}


/***************************************************************************
*                                                                          *
*   device_read:    This function is called whenever a process which       *
*                   already opened the device file attempts to read from   *
*                   it.                                                    *
*                                                                          *
*   Parameters:     inode:   we read from this inode...                    *
*                   file:    ... and from this file                        *
*                   buffer:  this will be filled with the read data        *
*                   length:  maximum buffer length! Don't write beyond.    *
*                                                                          *
*   Return Value:   number of bytes read into the buffer or errorcode:     *
*                   - return == length: all bytes were transferred.        *
*                   - return < length:  not all bytes were transferred.    *
*                   - return == 0:      End of File                        *
*                   - return < 0:       Error                              *
*                                                                          *
*   Revision:       0.4 (RS, 16.09.1999)                                   *
*                   0.3 (RS, 15.09.1999)                                   *
*                   0.2 (RS, 10.09.1999)                                   *
*                   0.1 (RS, 30.03.1999)                                   *                                         
*                                                                          *
***************************************************************************/

static int device_read(struct inode *inode,
                       struct file  *file,
                       char         *buffer,
                       int          length) 
{
  
  int copy;     /* number of bytes to copy from kernel buffer to userspace */

  /* Determine if there are bytes left in the kernel buffer for transfer.  */
  /* If not, check for rows which were not yet read from the camera and    */
  /* read the next one into the buffer, if there are any.                  */

  if (iBytesInBuffer == 0) {                  /* Buffer empty?             */

    if (iRowsLeft == 0) {                     /* all rows were read        */
    
      printk("LcCCD14: Buffer is empty.\n");
      return 0;

    } else {                                  /* There are unread rows!    */
                                              /* -> read them into buffer  */
      pu8RowBuffer = &u8RowBuffer[0];
      
      if (LcCCD14_Transfer_Row(&pu8RowBuffer,&SetupParams,&iBytesInBuffer)
          != LcCCD14_Successful) {
        printk("LcCCD14: Error transferring row from camera!\n");
      }	 

      pu8RowBuffer = &u8RowBuffer[0];         /* this was changed by the   */
      iRowsLeft--;                            /*  last command             */
      
    }
  }    
    
  /* Start transfer from Kernel space buffer to user space memory          */

  if (iBytesInBuffer >= length) { 
    copy = length;
  } else {
    copy = iBytesInBuffer; 
  }
  
  iDevice_Open++;                             /* for reentrance            */
  memcpy_tofs(buffer, pu8RowBuffer, copy);    /* copy kernel to user space */
  iDevice_Open--;   

  pu8RowBuffer   += copy;                     /* point to new reading pos. */
  iBytesInBuffer -= copy;                     /* these bytes are left      */

  return copy; 

}


/***************************************************************************
*                                                                          *
*   device_write:   This function is called when somebody tries to write   *
*                   into our device file - currently unsupported           *
*                                                                          *
*   Parameters:     inode, file, buffer, length                            *
*                                                                          *
*   Return Value:   ???                                                    *
*                                                                          *
*   Revision:       0.1 (30.03.1999)                                       *                                         
*                                                                          *
***************************************************************************/

/*
static int device_write(struct inode *inode,
                        struct file *file,
                        const char *buffer,
                        int length)
{
  printk ("LcCCD14: device_write - currently unsupported.\n");

  return -EINVAL;
}
*/


/* ----------------------------------------------------------------------- */
/*                                                                         */
/* ioctl() Declarations                                                    */
/*                                                                         */
/* ----------------------------------------------------------------------- */

static int device_ioctl(struct inode  *inode, 
                        struct file   *file,
		        unsigned int  cmd, 
			unsigned long arg) 
{

  unsigned int minor = MINOR(inode->i_rdev);
  int retval         = 0;
  enum   LcCCD14_ErrorEnum Status = LcCCD14_Undef;

  if (minor >= LP_NO)                   /* lp devices > 3 not supported    */
    return -ENODEV;

  switch ( cmd ) {
    case Lc_PRINT_PARAMS: printk("LcCCD14: Current Parameters:\n");
                          printk("         Integration time:     %d\n",
			          SetupParams.iIntegrationtime);
                          printk("         Desired Temperature:  %d\n",
			          SetupParams.iDesiredTemp);
                          printk("         Measured Temperature: %d\n",
                                  SetupParams.iMeasuredTemp);
                          printk("         Cooling:              %d\n",
                                  SetupParams.iCooling);
                          printk("         Resolution:           %d\n",
                                  SetupParams.iADResolution);
                          printk("         Binning:              %d\n",
                                  SetupParams.iBinning);
                          printk("         Sampling:             %d\n",
                                  SetupParams.iSampling);
                          printk("         Interlace:            %d\n",
                                  SetupParams.iInterl);
                          break;

    case Lc_SET_INTTIME:  if ((arg < 5) || (arg > 65534)) {
			    printk("LcCCD14: wrong integration time!\n");
			    return(-1);
			  }    
                          SetupParams.iIntegrationtime = (int) arg;
                          break;

    case Lc_SET_TEMP:     SetupParams.iDesiredTemp     = (int) arg;
                          break;

    case Lc_SET_COOLING:  if ((arg < 0) || (arg > 1)) {
                            printk("LcCCD14: wrong cooling status!\n");
			    return(-1);
			  }
                          SetupParams.iCooling         = (int) arg;
                          break;

    case Lc_SET_RES:      if ((arg < 0) || (arg > 1)) {
                            printk("LcCCD14: wrong AD resolution!\n");
			    return(-1);
			  }
                          SetupParams.iADResolution    = (int) arg;
                          break;

    case Lc_SET_BIN:      if ((arg < 0) || (arg > 1)) {
                            printk("LcCCD14: wrong binning status!\n");
			    return(-1);
			  }
                          SetupParams.iBinning         = (int) arg;
                          break;

    case Lc_SET_SAMP:     if ((arg < 0) || (arg > 1)) {
                            printk("LcCCD14: wrong sampling status!\n");
			    return(-1);
			  }
                          SetupParams.iSampling        = (int) arg;
                          break;

    case Lc_SET_INTER:    if ((arg < 0) || (arg > 1)) {
                            printk("LcCCD14: wrong interlace status!\n");
			    return(-1);
			  }
                          SetupParams.iInterl          = (int) arg;
                          break;

    case Lc_SETUP:        printk("LcCCD14: setup command: ");
                          Status = LcCCD14_Setup(&SetupParams);
                          switch(Status) {
			    case LcCCD14_Successful: 
			      printk("successful.\n");
			      break;
			    case LcCCD14_Timeout:
			      printk("timeout!\n");
			      break;
			    case LcCCD14_EchoErr:
			      printk("camera signal echo error!\n");
			      break;
			    case LcCCD14_WrongParam:
			      printk("parameter error!\n");
			      break;
			    default:
			      printk("unknown command!\n");
			      break;
			  }
                          break;

    case Lc_RESET:        printk("LcCCD14: reset command: ");
                          Status = LcCCD14_Reset();
			  if (Status != LcCCD14_Successful) {
			    printk("error! status: %d\n", Status);
			    return(-1);
			  }  else {
			    printk("successful.\n");
			    return(0);
			  }
		       
    case Lc_INTEGRATE:    printk("LcCCD14: starting integration: ");

                          Status =
			  LcCCD14_Integrate(SetupParams.iIntegrationtime /
			  1000.0);

			  switch(Status){

                            case LcCCD14_Successful: 
			    
			      printk("successful.\n");
			      /* images have always 290 rows               */
			      iRowsLeft = 290;    
			      /* data is not yet in the buffer!            */
			      iBytesInBuffer = 0; 
			      LcCCD14_Prepare_Transfer();
			      break;
			      
			    case LcCCD14_Timeout:
			      printk("timeout!\n");
			      break;
			    case LcCCD14_Userbreak:
			      printk("interrupted by user!\n");
			      break;
			    case LcCCD14_WrongParam:
			      printk("wrong parameters!\n");
			      break;
			    default:
			      break;			  
			  }

			  break;

    case Lc_DEBUG_REMOVE: printk("LcCCD14: DEBUG_REMOVE\n");
                          iDevice_Open = 0;
                          return(0);
		       
    default:              printk("LcCCD14: undefined IOCTL call\n");

  }

  return retval;

}


/* ----------------------------------------------------------------------- */
/*                                                                         */
/* Module Declarations                                                     */
/*                                                                         */
/* ----------------------------------------------------------------------- */

/* This structure will hold the functions to be called when                */
/* a process does something to the device we created. Since a pointer to   */
/* this structure is kept in the devices table, it can't be local to       */
/* init_module. NULL is for unimplemented functions.                       */ 

struct file_operations Fops = {
  NULL,                                 /* seek                            */
  device_read,                          /* read                            */
  NULL,                                 /* write                           */
  NULL,                                 /* readdir                         */
  NULL,                                 /* select                          */
  device_ioctl,                         /* ioctl                           */
  NULL,                                 /* mmap                            */
  device_open,                          /* open                            */
  device_release                        /* close                           */
};


/***************************************************************************
*                                                                          *
*   init_module:    Register the character device                          *
*                                                                          *
*   Parameters:     None.                                                  *
*                                                                          *
*   Return Value:   Error code from <asm/errno.h>:                         *
*									   *
*                   EIO: Driver is unable to reserve the desired major     *
*                        device number                                     *
*                   0:   Device was registered successfully.               *
*									   *
*   Revision:       1.0 (RS, 03.04.1999)                                   *                                                                           *
*                                                                          *
***************************************************************************/

int init_module()
{

  int ret = 0;

  /* Register the character device (at least try)                          */

  if (module_register_chrdev(DEVICE_MAJOR,DEVICE_NAME,&Fops)) {
    printk("LcCCD14: unable to get major %d\n", DEVICE_MAJOR);
    return -EIO;
  }

  printk ("LcCCD14: Device registered with major device number %d.\n",
    DEVICE_MAJOR);
  printk ("LcCCD14: A device file should be created with something like:\n");
  printk ("LcCCD14: mknod %s c %d 0\n", DEVICE_NAME, DEVICE_MAJOR);

  /* initialize global parameters                                          */
  printk ("LcCCD14: Initializing parameters.\n");
  SetupParams.iIntegrationtime = 5;
  SetupParams.iDesiredTemp     = 30;
  SetupParams.iMeasuredTemp    = 0;
  SetupParams.iCooling         = ausKuehlung;
  SetupParams.iADResolution    = AD4Bit;
  SetupParams.iBinning         = ausBinning;
  SetupParams.iSampling        = einfachAbt;
  SetupParams.iInterl          = ausInterlc;

  /* Request IO ports for lp1; the result can be observed in /proc/ioports */
  if (check_region (Port, 3)) {
    printk("LcCCD14: i/o port is locked by another process!\n");
    return -EBUSY; 
  }
  
  request_region(Port, 3, DEVICE_NAME); /* reserve 3 bytes                 */

  return ret;
}


/***************************************************************************
*                                                                          *
*   cleanup_module: Unregister the device                                  *
*                                                                          *
*   Parameters:     None.                                                  *
*                                                                          *
*   Return Value:   None.                                                  *
*									   *
*   Revision:       1.0 (RS, 03.04.1999)                                   *                                                                           *
*                                                                          *
***************************************************************************/

void cleanup_module()
{
  int ret = 0;

  /* release ports                                                         */
  release_region (Port, 3);
 
  /* Unregister the device                                                 */
  printk("LcCCD14: unregistering module: ");
  ret = module_unregister_chrdev(DEVICE_MAJOR, DEVICE_NAME);
  printk("return=%d\n", ret);
 
  /* If there's an error, report it                                        */ 
  if (ret < 0)
    printk("LcCCD14: Error in module_unregister_chrdev: %d\n", ret);
}

